# -*- coding: utf-8 -*-
"""
| Author: Alexandre CARRE 
| Created on: Jan 14, 2021
"""
import os
import pickle
import warnings
from typing import Union, List, Tuple, Optional

import numpy as np
import pandas as pd

from ..neurocombat.neurocombat import standardize_across_features
from ..neurocombat.neurocombat import find_parametric_adjustments
from ..neurocombat.neurocombat import fit_LS_model_and_find_priors
from ..neurocombat.neurocombat import find_non_eb_adjustments
from ..neurocombat.neurocombat import find_non_parametric_adjustments
# from ..neurocombat.neurocombat.neuroCombat import standardize_across_features
# from ..neurocombat.neurocombat.neuroCombat import find_parametric_adjustments
# from ..neurocombat.neurocombat.neuroCombat import fit_LS_model_and_find_priors
# from ..neurocombat.neurocombat.neuroCombat import find_non_eb_adjustments
# from ..neurocombat.neurocombat.neuroCombat import find_non_parametric_adjustments
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_is_fitted
from tqdm import tqdm

from .clustering import optimal_clustering
from .nifti import _compute_mask_files, flatten_nifti_files
from .utils import (
    get_column_index,
    one_hot_encoder,
    fix_columns,
    scaler_encoder,
    check_is_nii_exist,
    load_nifty_volume_as_array,
    save_to_nii,
    check_exist_vars,
)

__all__ = ["Combat", "AutoCombat", "ImageCombat"]


def _check_single_covariate_sample(df: pd.DataFrame, _vars: List) -> None:
    """
    Check if samples present single covariate
    :param df: a DataFrame
    :param _vars: List of columns name to check
    :raise ValueError if a covariate contain a unique sample
    """
    for _var in _vars:
        if 1 in df[_var].value_counts().tolist():
            raise ValueError(
                f"Combat ComScan requires more than one sample. "
                f"The following covariate contain a unique sample: {_var} "
            )


def _check_nans(df: pd.DataFrame) -> None:
    """
    Check if NaNs are present in dataframe

    :param df: a DataFrame
    :raise: ValueError if NaNs are present
    """
    if df.isnull().values.any():
        raise ValueError(
            "NaN values found on data. \n"
            "Combat can not work with NaN values,\n"
            " maybe drop samples or features columns"
            " containing these values"
        )


def _reorder_columns(all_columns: List[List]):
    """
    Reorder a list of list based on the total index
    :param all_columns: columns is a list of list
    :return: all_columns with new value as index
    """
    new_all_col_list = []
    i = 0
    for col in all_columns:
        col_list = []
        if col:
            for _ in col:
                col_list.append(i)
                i += 1
        new_all_col_list.append(col_list)
    return new_all_col_list


class Combat(BaseEstimator, TransformerMixin):
    """
    Harmonize/normalize features using Combat's parametric
    empirical Bayes framework

    .. seealso::
        `Fortin, Jean-Philippe, et al. "Harmonization of cortical thickness
        measurements across scanners and sites." Neuroimage 167 (2018): 104-1
        <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5845848>`__

    Parameters
    ----------
    features : Target features to be harmonized

    sites : Target variable for ComScan problems
        (e.g. acquisition sites or scanner).

    discrete_covariates : Target covariates which are
        categorical (e.g. male or female).

    continuous_covariates : Target covariates
        which are continuous (e.g. age).

    ref_site : Variable value (acquisition sites or scanner)
        to be used as reference for batch adjustment.
        Default is False.

    empirical_bayes : Performed empirical bayes.
        Default is True.

    parametric : Performed parametric adjustements.
        Default is True.

    mean_only : Adjust only the mean (no scaling).
        Default is False.

    return_only_features : Return only features.
        Default is False.

    raise_ref_site : raise when the reference site pass as arguments not exist, else set to no reference.
        Default is True.

    copy : Set to False to perform inplace row normalization and avoid a copy (if the input is already a numpy array).
        Default is True.

    Attributes
    ----------
    info_dict_fit\_ : dictionary that stores batch info of fitted data with:
         batch_levels, ref_level, n_batch, n_sample, sample_per_batch, batch_info

    stand_mean\_ : array-like
        Standardized mean

    var_pooled\_ : array-like
         Variance pooled

    mod_mean\_ : array-like
        Mod mean

    gamma_star\_ : array-like
        Adjustement gamma star

    delta_star\_ : array-like
        Adjustement delta star

    info_dict_transform\_ : dictionary that stores batch info of transformed data with
         batch_levels, ref_level, n_batch, n_sample, sample_per_batch, batch_info

    Examples
    --------


    Notes
    -----
    NaNs values are not treated.

    """

    def __init__(
        self,
        features: Union[List[str], List[int], str, int],
        sites: Union[str, int],
        discrete_covariates: Optional[Union[List[str], List[int], str, int]] = None,
        continuous_covariates: Optional[Union[List[str], List[int], str, int]] = None,
        ref_site: Optional[Union[str, int]] = None,
        empirical_bayes: bool = True,
        parametric: bool = True,
        mean_only: bool = False,
        return_only_features: bool = False,
        raise_ref_site: bool = True,
        copy: bool = True,
    ) -> None:

        self.features = features
        self.discrete_covariates = discrete_covariates
        self.continuous_covariates = continuous_covariates
        self.sites = sites
        self.ref_site = ref_site
        self.empirical_bayes = empirical_bayes
        self.parametric = parametric
        self.mean_only = mean_only
        self.return_only_features = return_only_features
        self.raise_ref_site = raise_ref_site
        self.copy = copy

    def __reset(self) -> None:
        """
        Reset internal data-dependent state of the combat fit, if necessary.
        __init__ parameters are not touched.
        """

        if hasattr(self, "info_dict_fit_"):
            del self.info_dict_fit_
            del self.stand_mean_
            del self.var_pooled_
            del self.gamma_star_
            del self.delta_star_

    def fit(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        *y: Optional[Union[np.ndarray, pd.DataFrame]],
    ) -> "Combat":
        """
        Compute the stand mean, var pooled, gamma star, delta star to be used for later adjusted data.

        Parameters
        ----------
        X : array-like or DataFrame of shape (n_samples, n_features).
            Requires the columns needed by the Combat().
            The data used to find adjustments.
        \*y : y in scikit learn: None
            Ignored.

        Returns
        -------
        self : object
            Fitted combat estimator.
        """
        self.__reset()

        all_columns = self._check_data(X)
        columns_needed = all_columns[0:4]

        (
            columns_features,
            columns_discrete_covariates,
            columns_continuous_covariates,
            columns_sites,
        ) = _reorder_columns(columns_needed)

        if isinstance(X, pd.DataFrame):
            X = X.to_numpy()

        X = self._validate_data(
            X[:, sum(columns_needed, [])], copy=self.copy, estimator=self
        )

        if self.ref_site is None:
            ref_level = None
        else:
            ref_indices = np.argwhere(X[:, columns_sites[0]] == self.ref_site).squeeze()
            if ref_indices.shape[0] == 0:
                if self.raise_ref_site:
                    raise ValueError(f"ref_site: {self.ref_site} not found")
                else:
                    ref_level = None
                    warnings.warn(
                        f"raise_ref_site is False and ref site not found. Setting to None",
                        UserWarning,
                        stacklevel=2,
                    )
            else:
                ref_level = np.int(X[ref_indices[0], columns_sites])

        # create dictionary that stores batch info
        (batch_levels, sample_per_batch) = np.unique(
            X[:, columns_sites], return_counts=True
        )
        self.info_dict_fit_ = {
            "batch_levels": batch_levels,
            "ref_level": ref_level,
            "n_batch": len(batch_levels),
            "n_sample": int(X.shape[0]),
            "sample_per_batch": sample_per_batch.astype("int"),
            "batch_info": [
                list(np.where(X[:, columns_sites] == idx)[0]) for idx in batch_levels
            ],
        }

        # create design matrix
        design = self._make_design_matrix(
            X=X,
            sites_col=columns_sites,
            ref_site=ref_level,
            discrete_covariates_col=columns_discrete_covariates,
            continuous_covariates_col=columns_continuous_covariates,
            fitting=True,
        )

        # standardize data across features
        s_data, self.stand_mean_, self.var_pooled_, self.mod_mean_ = (
            standardize_across_features(
                X=X[:, columns_features].T, design=design, info_dict=self.info_dict_fit_
            )
        )

        # fit L/S models and find priors
        LS_dict = fit_LS_model_and_find_priors(
            s_data=s_data,
            design=design,
            info_dict=self.info_dict_fit_,
            mean_only=self.mean_only,
        )

        # find parametric adjustments
        if self.empirical_bayes:
            if self.parametric:
                self.gamma_star_, self.delta_star_ = find_parametric_adjustments(
                    s_data=s_data,
                    LS=LS_dict,
                    info_dict=self.info_dict_fit_,
                    mean_only=self.mean_only,
                )
            else:
                self.gamma_star_, self.delta_star_ = find_non_parametric_adjustments(
                    s_data=s_data,
                    LS=LS_dict,
                    info_dict=self.info_dict_fit_,
                    mean_only=self.mean_only,
                )
        else:
            self.gamma_star_, self.delta_star_ = find_non_eb_adjustments(
                s_data=s_data, LS=LS_dict, info_dict=self.info_dict_fit_
            )

        return self

    def transform(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        """
        Scale features of X according to combat estimator.

        Parameters
        ----------
        X : array-like or DataFrame of shape (n_samples, n_features). Requires the columns needed by the Combat().
            Input data that will be transformed.

        Returns
        -------
        Xt : array-like of shape (n_samples, n_features)
            Transformed data.
        """
        check_is_fitted(self)

        all_columns = self._check_data(X, check_single_covariate=False)
        columns_needed = all_columns[0:4]

        # Warning: Here we reorder all columns
        (
            columns_features,
            columns_discrete_covariates,
            columns_continuous_covariates,
            columns_sites,
        ) = _reorder_columns(columns_needed)

        columns_df = None
        if isinstance(X, pd.DataFrame):
            columns_df = list(X.columns)
            if self.return_only_features:
                columns_df = list(X.columns[columns_needed[0]])
            X = X.to_numpy()

        # make a copy of original
        original_X = np.copy(X)

        X = self._validate_data(
            X[:, sum(columns_needed, [])], copy=self.copy, estimator=self
        )

        (batch_levels, sample_per_batch) = np.unique(
            X[:, columns_sites], return_counts=True
        )

        # Check all sites from new_data were seen
        if not all(
            site_name in self.info_dict_fit_["batch_levels"]
            for site_name in batch_levels
        ):
            raise ValueError(
                f"There is an unseen site during the fit method in the data"
            )

        # create dictionary that stores batch info
        self.info_dict_transform_ = {
            "batch_levels": batch_levels,
            "ref_level": self.ref_site,
            "n_batch": len(batch_levels),
            "n_sample": int(X.shape[0]),
            "sample_per_batch": sample_per_batch.astype("int"),
            "batch_info": [
                list(np.where(X[:, columns_sites] == idx)[0]) for idx in batch_levels
            ],
        }

        # create design matrix
        design = self._make_design_matrix(
            X=X,
            sites_col=columns_sites,
            ref_site=self.ref_site,
            discrete_covariates_col=columns_discrete_covariates,
            continuous_covariates_col=columns_continuous_covariates,
            fitting=False,
        )

        self.stand_mean_transform_ = np.repeat(
            self.stand_mean_[:, [0]], self.info_dict_transform_["n_sample"], axis=1
        )
        self.mod_mean_transform_ = np.repeat(
            self.mod_mean_[:, [0]], self.info_dict_transform_["n_sample"], axis=1
        )

        s_data = (
            X[:, columns_features].T
            - self.stand_mean_transform_
            - self.mod_mean_transform_
        ) / np.dot(
            np.sqrt(self.var_pooled_),
            np.ones((1, self.info_dict_transform_["n_sample"])),
        )

        # adjust data
        # ** obligatory to match the fit data for adjust_data (sample_per_batch, n_sample, batch_info)
        bayes_data = self._adjust_final_data(
            s_data=s_data,
            design=design,
            sample_per_batch=np.array(
                [
                    np.count_nonzero(X[:, columns_sites] == lvl)
                    for lvl in self.info_dict_fit_["batch_levels"]
                ]
            ),
            n_sample=self.info_dict_transform_["n_sample"],
            batch_info=[
                list(np.where(X[:, columns_sites] == idx)[0])
                for idx in self.info_dict_fit_["batch_levels"]
            ],
        ).T

        original_X[:, columns_needed[0]] = bayes_data

        if self.return_only_features:
            original_X = original_X[:, columns_needed[0]]

        if columns_df is not None:
            original_X = pd.DataFrame(original_X, columns=columns_df)

        return original_X

    def _check_data(
        self, X: Union[np.ndarray, pd.DataFrame], check_single_covariate: bool = True
    ) -> Tuple[List, List, List, List, List]:
        """
        Check that the input data array-like or DataFrame of shape (n_samples, n_features) have all the required
        format needed by the Combat()

        :param X: input data array-like or DataFrame of shape (n_samples, n_features)
        :param check_single_covariate: check single covariate
        :return: idx of columns_features, columns_discrete_covariates,
            columns_continuous_covariates, columns_sites, other_columns
        """
        (
            self.features,
            self.discrete_covariates,
            self.continuous_covariates,
            self.sites,
        ) = map(
            lambda x: [x] if isinstance(x, (str, int)) else x if x is not None else [],
            [
                self.features,
                self.discrete_covariates,
                self.continuous_covariates,
                self.sites,
            ],
        )

        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X.copy())

        columns_features = check_exist_vars(X, self.features)
        columns_sites = check_exist_vars(X, self.sites)
        columns_used = (columns_features, columns_sites)
        columns_discrete_covariates, columns_continuous_covariates = [], []
        if self.discrete_covariates:
            columns_discrete_covariates = check_exist_vars(X, self.discrete_covariates)
            _check_single_covariate_sample(X, self.discrete_covariates)
            columns_used += (columns_discrete_covariates,)
        if self.continuous_covariates:
            columns_continuous_covariates = check_exist_vars(
                X, self.continuous_covariates
            )
            _check_single_covariate_sample(X, self.continuous_covariates)
            columns_used += (columns_continuous_covariates,)

        unq, unq_cnt = np.unique(
            np.concatenate(
                (X.columns.get_indexer(X.columns), np.concatenate(columns_used))
            ),
            return_counts=True,
        )
        other_columns = unq[unq_cnt == 1]

        if check_single_covariate:
            _check_single_covariate_sample(X, self.sites)
            if self.discrete_covariates:
                _check_single_covariate_sample(X, self.discrete_covariates)
            if self.continuous_covariates:
                _check_single_covariate_sample(X, self.continuous_covariates)

        (
            columns_features,
            columns_discrete_covariates,
            columns_continuous_covariates,
            columns_sites,
            other_columns,
        ) = map(
            lambda x: x.tolist() if isinstance(x, np.ndarray) else x,
            [
                columns_features,
                columns_discrete_covariates,
                columns_continuous_covariates,
                columns_sites,
                other_columns,
            ],
        )

        _check_nans(
            X[
                self.features
                + self.discrete_covariates
                + self.continuous_covariates
                + self.sites
            ]
        )

        return (
            columns_features,
            columns_discrete_covariates,
            columns_continuous_covariates,
            columns_sites,
            other_columns,
        )

    def _make_design_matrix(
        self,
        X: np.ndarray,
        sites_col: int,
        ref_site: Optional[Union[str, int]] = None,
        discrete_covariates_col: Optional[List[int]] = None,
        continuous_covariates_col: Optional[List[int]] = None,
        fitting: bool = False,
    ) -> np.ndarray:
        """
        Method to create a design matrix that contain:
            - One-hot encoding of the sites [n_samples, n_sites]
            - One-hot encoding of each discrete covariates (removing
            the first column) [n_samples, (n_discrete_covivariate_names-1) * n_discrete_covariates]
            - Each continuous covariates

        :param X: array-like of shape (n_samples, n_features)
        :param sites_col: int, columns nb
        :param ref_site: ref sites in sites_col
        :param discrete_covariates_col: int, columns nb
        :param continuous_covariates_col: int, columns nb
        :param fitting: boolean, default is False
        :return design: the design matrix
        """
        design_list = []

        # Sites
        if fitting:
            self.site_encoder = OneHotEncoder(sparse_output=False)
            self.site_encoder.fit(X[:, sites_col])

        sites_design = self.site_encoder.transform(X[:, sites_col])
        if ref_site:
            if (
                len(self.site_encoder.categories_) > 1
                or not self.site_encoder.categories_
            ):
                raise NotImplementedError(
                    f"An error of encoding occurs: {self.site_encoder.categories_}"
                )

            sites_design[
                :, self.site_encoder.categories_[0].tolist().index(ref_site)
            ] = np.ones(sites_design.shape[0])

        design_list.append(sites_design)

        # Discrete covariates
        if discrete_covariates_col:
            if fitting:
                self.discrete_encoders = []
                for i in discrete_covariates_col:
                    discrete_encoder = OneHotEncoder(sparse_output=False)
                    discrete_encoder.fit(X[:, [i]])
                    self.discrete_encoders.append(discrete_encoder)

            for j, i in enumerate(discrete_covariates_col):
                discrete_encoder = self.discrete_encoders[j]
                discrete_covariate_one_hot = discrete_encoder.transform(X[:, [i]])
                discrete_covariate_design = discrete_covariate_one_hot[:, 1:]
                design_list.append(discrete_covariate_design)

        # Continuous covariates
        if continuous_covariates_col:
            design_list.append(X[:, continuous_covariates_col])
        design = np.hstack(design_list)

        return design

    def _adjust_final_data(
        self,
        s_data: np.ndarray,
        design: np.ndarray,
        sample_per_batch: np.ndarray,
        n_sample: int,
        batch_info: np.ndarray,
    ) -> np.ndarray:
        """
        Adjust final data

        :param s_data: array-like standardized data
        :param design: array-like design matrix
        :param sample_per_batch: number of sample per batch
        :param n_sample: total number of sample
        :param batch_info: batch info is batch index for n_batch
        :return: bayes data
        """
        bayes_data = s_data
        batch_design = design[:, : self.info_dict_fit_["n_batch"]]

        for j, batch_idxs in enumerate(batch_info):
            if not batch_idxs:
                continue
            dsq = np.sqrt(self.delta_star_[j, :])
            dsq = dsq.reshape((len(dsq), 1))
            denom = np.dot(dsq, np.ones((1, sample_per_batch[j])))
            numer = np.array(
                bayes_data[:, batch_idxs]
                - np.dot(batch_design[batch_idxs, :], self.gamma_star_).T
            )

            bayes_data[:, batch_idxs] = numer / denom

        vpsq = np.sqrt(self.var_pooled_).reshape((len(self.var_pooled_), 1))
        bayes_data = (
            bayes_data * np.dot(vpsq, np.ones((1, n_sample)))
            + self.stand_mean_transform_
            + self.mod_mean_transform_
        )

        return bayes_data

    def save_fit(self, filepath: str) -> None:
        """
        save a fitted model attribute :py:obj:`info_dict_fit_`, :py:obj:`stand_mean_`, :py:obj:`var_pooled_`,
        :py:obj:`gamma_star_`, :py:obj:`delta_star_`

        :param filepath: filepath were to save. if no extension .pkl will add it
        """
        if not os.path.exists(filepath):
            os.makedirs(filepath, exist_ok=True)

        if not filepath.endswith(".pkl"):
            filepath += ".pkl"

        attrs_to_save = {
            k: v
            for k, v in vars(self).items()
            if k.endswith("_") and not k.startswith("__")
        }

        with open(filepath, "wb") as output:  # Overwrites any existing file.
            pickle.dump(attrs_to_save, output, pickle.HIGHEST_PROTOCOL)

    def load_fit(self, filepath: str) -> None:
        """
        load a fitted model attribute :py:obj:`info_dict_fit_`, :py:obj:`stand_mean_`, :py:obj:`var_pooled_`,
        :py:obj:`gamma_star_`, :py:obj:`delta_star_`

        :param filepath: filepath of the pkl file to load
        """
        with open(filepath, "rb") as pickle_file:  # Overwrites any existing file.
            loaded_pickle = pickle.load(pickle_file)
        for k, v in loaded_pickle.items():
            setattr(self, k, v)


class AutoCombat(Combat):
    """
    Harmonize/normalize features using Combat's parametric empirical Bayes framework.

    Combat need to have well-known acquisition sites or scanner to harmonize features.
    It is sometimes difficult to define an imaging acquisition site if on two sites imaging parameters
    can be really similar. ComScan gives the possibility to automatically determine the number of sites
    and their association based on imaging features (e.g. dicom tags) by clustering.
    Thus ComScan can be used on data not seen during training because it can predict which imager best matches
    the one it has seen during training.

    .. seealso::
        `Fortin, Jean-Philippe, et al. "Harmonization of cortical thickness
        measurements across scanners and sites." Neuroimage 167 (2018): 104-1
        <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5845848>`__

    Parameters
    ----------
    features : Target features to be harmonized.

    sites_features : Target variable for define (acquisition sites or scanner) by clustering.

    sites : Target variable for ComScan problems (e.g. acquisition sites or scanner).
        This argument is Optional. If this argument is provided will run traditional ComBat else AutoCombat.
        In this case args: sites_features, size_min, method, scaler_clustering, discrete_cluster_features,
        continuous_cluster_features, threshold_missing_sites_features, drop_site_columns
        are unused.

    size_min : Constraint of the minimum size of site for clustering.

    metric : "distortion", "silhouette" or "calinski_harabasz".
        Metric to define the optimal number of cluster. Default: distortion.

    use_ref_site : Use a ref site to be used as reference for batch adjustment. The ref site used is the cluster
        with the minimal inertia. i.e minimizing within-cluster sum-of-squares.

    scaler_clustering : Scaler to use for continuous site features. Need to be a scikit learn scaler.
        Default is :py:class:`StandardScaler()`.

    discrete_cluster_features : Target sites_features which are categorical to one-hot (e.g. ManufacturerModelName).

    continuous_cluster_features : Target sites_features which are continuous to scale (e.g. EchoTime).

    features_reduction : Method for reduction of the embedded space with n_components. Can be 'pca' or 'umap'.
        Default is None.

    n_components : Dimension of the embedded space for features reduction.
        Default is 2.

    threshold_missing_sites_features : Threshold of acceptable missing features for sites features clustering.
        25 specify that 75% of all samples need to have this features.
        Default is 25.

    drop_site_columns : Drop sites columns find by clustering in return.
        Default is False.

    discrete_combat_covariates : Target covariates which are categorical (e.g. male or female).

    continuous_combat_covariates : Target covariates which are continuous (e.g. age).

    empirical_bayes : Performed empirical bayes.
        Default is True.

    parametric : Performed parametric adjustements.
        Default is True.

    mean_only : Adjust only the mean (no scaling)
        Default is False.

    return_only_features : Return only features.
        Default is False.

    n_jobs: The number of jobs to use for the computation. This works by computing each of the n_init runs in parallel.
        If -1 all CPUs are used. If 1 is given, no parallel computing code is
        used at all, which is useful for debugging. For n_jobs below -1,
        (n_cpus + 1 + n_jobs) are used. Thus for n_jobs = -2, all CPUs but one
        are used.
        Default is 1.


    random_state : int, RandomState instance or None, optional, default: 123
        If int, random_state is the seed used by the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    copy : Set to False to perform inplace row normalization and avoid a copy (if the input is already a numpy array).
        Default is True.

    Attributes
    ----------

    cls\_ : clustering classifier object

    info_clustering\_ : Dictionary that stores info of clustering from sites_features with cluster_nb, labels, ref_label
        wicss_clusters, best_wicss_cluster

    cls_feature_reduction\_ : feature reduction object

    clustering_data_features_mean\_ : dict of mean for clustering data (use for imputation)

    X_hat\_ : array after fit

    clustering_data_features\_ : column features for clustering from train (after encoding + scaling)

    clustering_data_discrete_features\_: column features for clustering after one-hot encoding

    dict_cls_fitted: dict of columns of fitted cls used for fitted clustering data

    Examples
    --------
    >>> data = pd.DataFrame([{"features_1": 0.97, "site_features_0": 2, "site_features_1": 0},
    >>> {"features_1": 1.35, "site_features_0": 1.01, "site_features_1": 1},
    >>> {"features_1": 1.43, "site_features_0": 1.09, "site_features_1": 1},
    >>> {"features_1": 0.85, "site_features_0": 2.3, "site_features_1": 0}])

    >>> auto_combat = AutoCombat(features=["features_1"], sites_features=["site_features_0", "site_features_1"],
    >>> continuous_cluster_features=["site_features_0", "site_features_1"], size_min=2))
    >>> print(auto_combat.fit(data))
    AutoCombat(continuous_cluster_features=['site_features_0', 'site_features_1'],
           discrete_cluster_features=[], features=['features_1'],
           sites=['sites'],
           sites_features=['site_features_0', 'site_features_1'], size_min=2))

    Notes
    -----
    NaNs values are not treated.

    Warning
    _______
    Be sure to have the same sites features between fit and transform. The choice has not been to imposed
    an entry format to check a colum name or a slice.

    """

    def __init__(
        self,
        features: Union[List[str], List[int], str, int],
        sites_features: Union[List[str], List[int], str, int] = None,
        sites: Optional[Union[str, int]] = None,
        size_min: int = 10,
        metric: str = "distortion",
        use_ref_site: bool = False,
        scaler_clustering=StandardScaler(),
        discrete_cluster_features: Optional[
            Union[List[str], List[int], str, int]
        ] = None,
        continuous_cluster_features: Optional[
            Union[List[str], List[int], str, int]
        ] = None,
        features_reduction: Optional[str] = None,
        n_components: int = 2,
        threshold_missing_sites_features=25,
        drop_site_columns: bool = False,
        discrete_combat_covariates: Optional[
            Union[List[str], List[int], str, int]
        ] = None,
        continuous_combat_covariates: Optional[
            Union[List[str], List[int], str, int]
        ] = None,
        empirical_bayes: bool = True,
        parametric: bool = True,
        mean_only: bool = False,
        return_only_features: bool = False,
        n_jobs: int = 1,
        random_state: Union[int, None] = 123,
        copy: bool = True,
    ) -> None:
        super().__init__(
            features=features,
            sites=sites,
            discrete_covariates=discrete_combat_covariates,
            continuous_covariates=continuous_combat_covariates,
            empirical_bayes=empirical_bayes,
            parametric=parametric,
            mean_only=mean_only,
        )

        if not 0 < threshold_missing_sites_features < 100:
            raise ValueError(
                "threshold_missing_sites_features need to be comprise between 0 and 100"
            )

        if (sites_features is None and sites is None) or (
            sites_features is not None and sites is not None
        ):
            raise ValueError(
                "one of sites_features or sites must be provided. If sites will run traditional ComBat"
            )

        self.features = features
        self.sites_features = sites_features
        self.size_min = size_min
        self.metric = metric
        self.use_ref_site = use_ref_site
        self.scaler_clustering = scaler_clustering
        self.discrete_cluster_features = discrete_cluster_features
        self.continuous_cluster_features = continuous_cluster_features
        self.features_reduction = features_reduction
        self.n_components = n_components
        self.threshold_missing_sites_features = threshold_missing_sites_features
        self.drop_site_columns = drop_site_columns
        self.discrete_combat_covariates = discrete_combat_covariates
        self.continuous_combat_covariates = continuous_combat_covariates
        self.empirical_bayes = empirical_bayes
        self.parametric = parametric
        self.mean_only = mean_only
        self.return_only_features = return_only_features
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.copy = copy

    def __reset(self) -> None:
        """
        Reset internal data-dependent state of the AutoCombat fit, if necessary.
        __init__ parameters are not touched.
        """

        if hasattr(
            self,
            "cls_",
        ):
            del self.cls_
            del self.cls_feature_reduction_
            del self.clustering_data_features_
            del self.clustering_data_features_mean_
            del self.info_clustering_
            del self.X_hat_
            del self.dict_cls_fitted

        if hasattr(
            self,
            "clustering_data_discrete_features_",
        ):
            del self.clustering_data_discrete_features_

    def fit(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        *y: Optional[Union[np.ndarray, pd.DataFrame]],
    ) -> "AutoCombat":
        """
        Compute sites, ref_site using clustering. Then compute the stand mean, var pooled, gamma star, delta star
        to be used for later adjusted data from Combat.

        Parameters
        ----------
        X : array-like or DataFrame of shape (n_samples, n_features).
            Requires the columns needed by the ComScan().
            The data used to find adjustments.
        \*y : y in scikit learn: None
            Ignored.

        Returns
        -------
        self : object
            Fitted ComScan estimator.
        """
        self.__reset()

        if self.sites_features is not None:

            (
                clustering_data,
                columns_clustering_features,
                columns_discrete_cluster_features,
                columns_continuous_cluster_features,
            ) = self._check_data_cluster(X)

            # get clustering data columns
            self.clustering_data_features_ = self.sites_features

            (
                self.cls_,
                self.cls_feature_reduction_,
                cluster_nb,
                labels,
                ref_label,
                wicss_clusters,
                best_wicss_cluster,
                _,
                self.X_hat_,
                mu,
            ) = optimal_clustering(
                X=clustering_data,
                size_min=self.size_min,
                metric=self.metric,
                features_reduction=self.features_reduction,
                n_jobs=self.n_jobs,
                random_state=self.random_state,
            )

            self.clustering_data_features_mean_ = (
                pd.DataFrame(mu, columns=list(clustering_data.columns))
                .iloc[0]
                .to_dict()
            )

            self.info_clustering_ = {
                "cluster_nb": cluster_nb,
                "labels": labels,
                "ref_label": ref_label,
                "wicss_clusters": wicss_clusters,
                "best_wicss_cluster": best_wicss_cluster,
            }

            if cluster_nb == 1:
                raise ValueError(
                    "Combat can not run. "
                    "Only one acquisition site found from sites features. "
                    "Are the data really different or from different acquisition site?"
                )

            # add sites columns
            X = self._add_sites(X, labels)

            if self.use_ref_site:
                self.ref_site = ref_label

        super().fit(X, *y)

        return self

    def transform(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        """
        Scale features of X according to combat estimator.

        Parameters
        ----------
        X : array-like or DataFrame of shape (n_samples, n_features). Requires the columns needed by the Combat().
            Input data that will be transformed.

        Returns
        -------
        Xt : array-like of shape (n_samples, n_features)
            Transformed data.
        """
        check_is_fitted(self)

        if self.sites_features is not None:
            (
                clustering_data,
                columns_clustering_features,
                columns_discrete_cluster_features,
                columns_continuous_cluster_features,
            ) = self._check_data_cluster(X)

            # check for NaN for clustering data
            clustering_data.fillna(
                value=self.clustering_data_features_mean_, inplace=True
            )

            if self.features_reduction is not None:
                clustering_data = self.cls_feature_reduction_.transform(clustering_data)

            # get labels for sites
            # tricky to avoid error with this verification:
            # size_min * n_clusters > n_samples: n_sample = clustering_data.shape[0]
            self.cls_.n_clusters = (
                1  # can be set to 1 because cls_.predict don't use this args
            )
            # don't
            labels = self.cls_.predict(clustering_data, size_min=None, size_max=None)

            # add sites columns
            X = self._add_sites(X, labels)

        X = super().transform(X)

        if self.sites_features is not None:
            # drop added sites columns
            if self.drop_site_columns and not self.return_only_features:
                if isinstance(X, pd.DataFrame):
                    X = X.iloc[:, :-1]
                else:
                    X = X[:, :-1]

        return X

    def _check_data_cluster(
        self, X: Union[np.ndarray, pd.DataFrame]
    ) -> Tuple[pd.DataFrame, List, List, List]:
        """
        Check that the input data array-like or DataFrame of shape (n_samples, n_features) have all the required
        format needed by the :py:class:`Combat()`

        :param X: input data array-like or DataFrame of shape (n_samples, n_features)
        :return: idx of: columns_clustering_features
        """

        if not self.sites_features:
            raise ValueError("sites_features is empty")

        (
            self.sites_features,
            self.discrete_cluster_features,
            self.continuous_cluster_features,
        ) = map(
            lambda x: [x] if isinstance(x, (str, int)) else x if x is not None else [],
            [
                self.sites_features,
                self.discrete_cluster_features,
                self.continuous_cluster_features,
            ],
        )

        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X.copy())

        # check that all clustering_data_features_ from fit are in transform
        if hasattr(self, "clustering_data_features_"):
            check_exist_vars(X, self.clustering_data_features_)

        columns_clustering_features = check_exist_vars(X, self.sites_features)

        columns_used = (columns_clustering_features,)
        columns_discrete_cluster_features, columns_continuous_cluster_features = [], []
        if self.discrete_cluster_features:
            columns_discrete_cluster_features = check_exist_vars(
                X, self.discrete_cluster_features
            )
            columns_used += (columns_discrete_cluster_features,)
        if self.continuous_cluster_features:
            columns_continuous_cluster_features = check_exist_vars(
                X, self.continuous_cluster_features
            )
            columns_used += (columns_continuous_cluster_features,)

        unq, unq_cnt = np.unique(
            np.concatenate((columns_clustering_features, np.concatenate(columns_used))),
            return_counts=True,
        )
        other_columns = unq[unq_cnt == 1]

        if other_columns.size > 0:
            warnings.warn(
                f"Some columns: {X.columns[other_columns].tolist()} are not specified as discrete or continuous. "
                f"Clustering will interpret this data as raw.",
                UserWarning,
                stacklevel=2,
            )

        percent_missing = (
            X.iloc[:, columns_clustering_features].isnull().sum() * 100 / len(X)
        )
        percent_missing = percent_missing.to_dict()

        if not hasattr(self, "clustering_data_features_"):
            self.sites_features_removed_ = []
            for features, val in percent_missing.items():
                if val > self.threshold_missing_sites_features:
                    warnings.warn(
                        f"sites_features: {features} removed because more than "
                        f"{self.threshold_missing_sites_features}% of missing data",
                        UserWarning,
                        stacklevel=2,
                    )
                    self.sites_features_removed_.append(features)
                    (
                        columns_clustering_features,
                        columns_discrete_cluster_features,
                        columns_continuous_cluster_features,
                    ) = list(
                        map(
                            lambda x: np.delete(
                                x, np.where(x == get_column_index(X, [features])[0])
                            ),
                            (
                                columns_clustering_features,
                                columns_discrete_cluster_features,
                                columns_continuous_cluster_features,
                            ),
                        )
                    )

        (
            columns_clustering_features,
            columns_discrete_cluster_features,
            columns_continuous_cluster_features,
        ) = map(
            lambda x: x.tolist() if isinstance(x, np.ndarray) else x,
            [
                columns_clustering_features,
                columns_discrete_cluster_features,
                columns_continuous_cluster_features,
            ],
        )

        # remove same features has in train
        if self.sites_features_removed_ and hasattr(self, "clustering_data_features_"):
            for feature_to_remove in self.sites_features_removed_:
                list(
                    map(
                        lambda x: (
                            x.remove(feature_to_remove) if feature_to_remove in x else x
                        ),
                        (
                            columns_clustering_features,
                            columns_discrete_cluster_features,
                            columns_continuous_cluster_features,
                        ),
                    )
                )

        clustering_data_discrete, clustering_data_continuous = pd.DataFrame(
            []
        ), pd.DataFrame([])
        if columns_discrete_cluster_features:
            clustering_data_discrete = one_hot_encoder(
                df=X.iloc[:, columns_discrete_cluster_features],
                columns=list(X.iloc[:, columns_discrete_cluster_features].columns),
                dummy_na=True,
                add_nan_columns=True,
            )

            # pure transform
            if hasattr(self, "clustering_data_discrete_features_"):
                clustering_data_discrete = fix_columns(
                    df=clustering_data_discrete,
                    columns=self.clustering_data_discrete_features_,
                    inplace=False,
                    extra_nans=True,
                )
            else:
                self.clustering_data_discrete_features_ = list(
                    clustering_data_discrete.columns
                )

        if columns_continuous_cluster_features:

            # pure transform
            if hasattr(self, "clustering_data_features_"):
                clustering_data_continuous = pd.DataFrame(
                    index=list(X.index), columns=list(self.dict_cls_fitted.keys())
                )
                for col, scal in self.dict_cls_fitted.items():
                    try:
                        clustering_data_continuous[col] = scal.fit_transform(X[col])
                    except ValueError:
                        clustering_data_continuous[col] = scal.fit_transform(
                            pd.DataFrame(X[col])
                        )
            else:
                clustering_data_continuous, self.dict_cls_fitted = scaler_encoder(
                    df=X.iloc[:, columns_continuous_cluster_features],
                    columns=list(
                        X.iloc[:, columns_continuous_cluster_features].columns
                    ),
                    scaler=self.scaler_clustering,
                )

        clustering_data = pd.concat(
            [clustering_data_discrete, clustering_data_continuous], axis=1
        )

        return (
            clustering_data,
            columns_clustering_features,
            columns_discrete_cluster_features,
            columns_continuous_cluster_features,
        )

    def _add_sites(
        self, X: Union[np.ndarray, pd.DataFrame], labels: Union[np.ndarray, List]
    ) -> Union[np.ndarray, pd.DataFrame]:
        """
        Add sites find by clustering to X

        :param X: input data array-like or DataFrame of shape (n_samples, n_features)
        :param labels: labels of sites find by clustering
        """

        if self.copy:
            X = X.copy(deep=True)

        sites = None
        if isinstance(X, pd.DataFrame):
            X["sites"] = labels
            sites = "sites"
        elif isinstance(X, np.ndarray):
            X = np.c_[X, labels]
            sites = X.shape[1]

        self.sites = sites

        return X


class ImageCombat(AutoCombat):
    """
    Harmonize/normalize features using Combat's parametric empirical Bayes framework directly on image.

    ImageCombat allow the possibility to Harmonize/normalize a set of NIFTI images.
    All images must have the same dimensions and orientation. A common mask is created based on an heuristic
    proposed by T.Nichols. Images are then vectorizing for ComScan.
    ImageCombat allows the possibily to use Combat (well-defined site) or AutoCombat (clustering for sites finding)

    .. seealso::
        `Fortin, Jean-Philippe, et al. "Harmonization of cortical thickness
        measurements across scanners and sites." Neuroimage 167 (2018): 104-1
        <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5845848>`__

    Parameters
    ----------
    image_path : image_path of nifti files.

    sites_features : Target variable for define (acquisition sites or scanner) by clustering.

    sites : Target variable for ComScan problems (e.g. acquisition sites or scanner).
        This argument is Optional. If this argument is provided will run traditional ComBat.
        In this case args: sites_features, size_min, method, scaler_clustering, discrete_cluster_features,
        continuous_cluster_features, threshold_missing_sites_features, drop_site_columns are unused.

    size_min : Constraint of the minimum size of site for clustering.

    method : "silhouette" or "elbow". Method to define the optimal number of cluster.
        Default is silhouette.

    use_ref_site: Use a ref site to be used as reference for batch adjustment. The ref site used is the cluster
        with the minimal inertia. i.e minimizing within-cluster sum-of-squares.
        Default is False.

    scaler_clustering: Scaler to use for continuous site features. Need to be a scikit learn scaler.
        Default is :py:class:`StandardScaler()`.

    discrete_cluster_features: Target sites_features which are categorical to one-hot (e.g. ManufacturerModelName).

    continuous_cluster_features: Target sites_features which are continuous to scale (e.g. EchoTime).

    features_reduction: Method for reduction of the embedded space with n_components. Can be 'pca' or 'umap'.
        Default is None.

    n_components: Dimension of the embedded space for features reduction.
        Default is 2.

    threshold_missing_sites_features: Threshold of acceptable missing features for sites features clustering.
        25 specify that 75% of all samples need to have this features.
        Default is 25.

    drop_site_columns: Drop sites columns find by clustering in return.

    discrete_combat_covariates : Target covariates which are categorical (e.g. male or female).

    continuous_combat_covariates : Target covariates which are continuous (e.g. age).

    empirical_bayes : Performed empirical bayes.
        Default is True.

    parametric : Performed parametric adjustements.
        Default is True.

    mean_only : Adjust only the mean (no scaling)
        Default is False.

    random_state: int, RandomState instance or None, optional, default: 123
        If int, random_state is the seed used by the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    copy : Set to False to perform inplace row normalization and avoid a copy (if the input is already a numpy array).
        Default is True.

    Attributes
    ----------

    mask\_ : array-like of the common brain mask

    flattened_array\_ : flattened array of all the training set


    Notes
    -----
    NaNs values are not treated.

    """

    def __init__(
        self,
        image_path: Union[str, int],
        sites_features: Union[List[str], List[int], str, int] = None,
        sites: Union[str, int] = None,
        save_path_fit: str = "fit_data",
        save_path_transform: str = "transform_data",
        size_min: int = 10,
        method: str = "silhouette",
        use_ref_site: bool = False,
        scaler_clustering=StandardScaler(),
        discrete_cluster_features: Optional[
            Union[List[str], List[int], str, int]
        ] = None,
        continuous_cluster_features: Optional[
            Union[List[str], List[int], str, int]
        ] = None,
        features_reduction: Optional[str] = None,
        n_components: int = 2,
        threshold_missing_sites_features=25,
        drop_site_columns: bool = True,
        discrete_combat_covariates: Optional[
            Union[List[str], List[int], str, int]
        ] = None,
        continuous_combat_covariates: Optional[
            Union[List[str], List[int], str, int]
        ] = None,
        empirical_bayes: bool = True,
        parametric: bool = True,
        mean_only: bool = False,
        random_state: Union[int, None] = 123,
        flattened_dtype: Optional[np.dtype] = np.float16,
        output_dtype: Optional[np.dtype] = np.float32,
        copy: bool = True,
    ) -> None:
        super().__init__(
            features=[],
            sites_features=sites_features,
            sites=sites,
            size_min=size_min,
            method=method,
            use_ref_site=use_ref_site,
            scaler_clustering=scaler_clustering,
            discrete_cluster_features=discrete_cluster_features,
            continuous_cluster_features=continuous_cluster_features,
            features_reduction=features_reduction,
            n_components=n_components,
            threshold_missing_sites_features=threshold_missing_sites_features,
            drop_site_columns=drop_site_columns,
            discrete_combat_covariates=discrete_combat_covariates,
            continuous_combat_covariates=continuous_combat_covariates,
            empirical_bayes=empirical_bayes,
            parametric=parametric,
            mean_only=mean_only,
            random_state=random_state,
            copy=copy,
        )

        self.image_path = image_path
        self.save_path_fit = save_path_fit
        self.save_path_transform = save_path_transform
        self.flattened_dtype = flattened_dtype
        self.output_dtype = output_dtype

        # For sequential transform
        self.sequential_transform = True

    def __reset(self) -> None:
        """
        Reset internal data-dependent state of the ImageCombat fit, if necessary.
        __init__ parameters are not touched.
        """

        if hasattr(
            self,
            "mask_",
        ):
            del self.mask_
            del self.flattened_array_

    def fit(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        *y: Optional[Union[np.ndarray, pd.DataFrame]],
    ) -> "ImageCombat":

        _, list_image_path = self._check_image_path(X)

        # create common brain mask
        self.mask_, _ = _compute_mask_files(
            input_path=list_image_path,
            output_path=os.path.join(self.save_path_fit, "common_mask.nii.gz"),
            return_mean=False,
        )

        # flatten nifti files
        self.flattened_array_ = flatten_nifti_files(
            input_path=list_image_path,
            mask=self.mask_,
            output_flattened_array_path=os.path.join(
                self.save_path_fit, "flattened_array"
            ),
            dtype=self.flattened_dtype,
            save=True,
            compress_save=True,
        )

        X, _ = self._add_voxels_as_features(X.copy(deep=True), self.flattened_array_)

        # run AutoCombat fit
        super().fit(X, *y)

    def transform(self, X: Union[np.ndarray, pd.DataFrame]) -> None:
        """
        Scale image according to combat estimator and save it.

        Parameters
        ----------
        X : array-like or DataFrame of shape (n_samples, n_features). Requires the columns needed by the ImageCombat().
            Input data that will be transformed.

        Returns
        -------
            None, save transformed image
        """
        check_is_fitted(self)

        _, list_image_path = self._check_image_path(X)

        logical_mask = self.mask_ == 1  # force the mask to be logical type
        n_voxels_flattened = np.sum(logical_mask)

        # to avoid OOM, apply sequentially the transform for image path
        for i, image_path in enumerate(tqdm(list_image_path, desc="Transform image")):
            to_convert = X.loc[[i]] if isinstance(X, pd.DataFrame) else X[[i]]
            flattened_array = np.zeros((1, n_voxels_flattened)).astype(
                self.flattened_dtype
            )
            image_arr, header = load_nifty_volume_as_array(image_path)
            flattened_array[0, :] = image_arr[logical_mask]
            to_convert, features_columns = self._add_voxels_as_features(
                to_convert.copy(deep=self.copy), flattened_array
            )
            # run AutoCombat transform
            adjusted_array = super().transform(to_convert)
            adjusted_array = adjusted_array[:, features_columns]
            #
            nifti_out = logical_mask.copy().astype(self.output_dtype)
            nifti_out[logical_mask] = adjusted_array[0, :]
            save_to_nii(
                im=nifti_out,
                header=header,
                output_dir=self.save_path_transform,
                filename=os.path.basename(image_path),
            )

    def _check_image_path(
        self, X: Union[np.ndarray, pd.DataFrame]
    ) -> Tuple[List, List]:

        self.image_path = (
            [self.image_path]
            if isinstance(self.image_path, (str, int))
            else self.image_path
        )

        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X.copy())

        column_image_path = check_exist_vars(X, self.image_path).tolist()

        if X[self.image_path[0]].isnull().values.any():
            raise ValueError("An image path is missing in rows")

        # check nii file exit
        X[self.image_path[0]].apply(lambda x: check_is_nii_exist(x))

        list_image_path = X[self.image_path[0]].to_list()

        return column_image_path, list_image_path

    def _add_voxels_as_features(
        self, X: Union[np.ndarray, pd.DataFrame], flattened_array: np.ndarray
    ):
        """
        Add voxels as features and drop the image_path columns
        :param X: Initial input of ImageCombat
        :param flattened_array: flattened array
        :return: X concatenate with the flattened array
        """

        init_columns = X.shape[1]
        end_columns = init_columns + flattened_array.shape[1]
        features_columns = list(range(init_columns, end_columns))

        if isinstance(X, pd.DataFrame):
            X[self.image_path] = (
                0.0  # create null columns for image path, allows to keep order
            )
            # parameters of columns parameters (because string type)
            X = pd.concat(
                [X.reset_index(drop=True), pd.DataFrame(flattened_array)], axis=1
            )
            self.features = list(range(flattened_array.shape[1]))
            if (np.unique(X.columns.to_list(), return_counts=True)[1] > 1).any():
                raise ValueError(
                    f"There is a column in the dataframe with the name in the range: "
                    f"{features_columns[0]} ... {features_columns[-1]} "
                )
        elif isinstance(X, np.ndarray):
            X[:, 1] = 0.0
            X = np.c_[X, flattened_array]
            self.features = features_columns

        return X, features_columns



# -*- coding: utf-8 -*-
"""
| Author: Alexandre CARRE 
| Created on: Jan 14, 2021
"""
import os
import re
from typing import List, Tuple, Union, Optional, Callable

import numpy as np
from nipy.labs.mask import compute_mask_files
from tqdm import tqdm

from .utils import load_nifty_volume_as_array
from .utils import mat_to_bytes


def _compute_mask_files(
    input_path: List[str],
    output_path: Optional[str] = None,
    return_mean: bool = False,
    m: float = 0.2,
    M: float = 0.9,
    cc: int = 1,
    exclude_zeros: bool = False,
    opening: int = 2,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """
    Wrap the nipy compute mask files function.


    Compute a mask file from MRI nifti file(s)

    Compute and write the mask of an image based on the grey level
    This is based on an heuristic proposed by T.Nichols:
    find the least dense point of the histogram, between fractions
    m and M of the total image histogram.

    In case of failure, it is usually advisable to increase m.

    :param input_path: string. list of filenames (3D).
    :param output_path: string or None, optional
        path to save the output nifti image (if not None).
    :param return_mean: boolean, optional
        if True, and output_filename is None, return the mean image also, as
        a 3D array (2nd return argument).
    :param m: float, optional
        lower fraction of the histogram to be discarded.
    :param M: float, optional
        upper fraction of the histogram to be discarded.
    :param cc: boolean, optional
        if cc is True, only the largest connect component is kept.
    :param exclude_zeros: boolean, optional
        Consider zeros as missing values for the computation of the
        threshold. This option is useful if the images have been
        resliced with a large padding of zeros.
    :param opening: int, optional
        Size of the morphological opening performed as post-processing
    :returns:
        - mask:
          3D boolean array. The brain mask
        - mean_image:
          3d ndarray, optional
          The main of all the images used to estimate the mask. Only provided if `return_mean` is True.
    """

    if not os.path.exists(os.path.dirname(output_path)):
        os.makedirs(os.path.dirname(output_path), exist_ok=True)

    output = compute_mask_files(
        input_filename=input_path,
        output_filename=output_path,
        return_mean=return_mean,
        m=m,
        M=M,
        cc=cc,
        exclude_zeros=exclude_zeros,
        opening=opening,
    )

    if return_mean:
        mask, mean_volume = output[0], output[1]
    else:
        mask, mean_volume = output, None

    # nibabel array is x,y,z and sitk is z,y,x -> convert nibabel to sitk
    mask = np.swapaxes(mask, 0, 2)

    return mask, mean_volume


def flatten_nifti_files(
    input_path: List[str],
    mask: Union[str, np.ndarray],
    output_flattened_array_path: str = "flattened_array",
    dtype: [np.dtype, Callable] = np.float16,
    save: bool = True,
    compress_save: bool = True,
):
    """
    Flattened list of nifti files to a flattened array [n_images, n_masked_voxels] and save to .npy or .npz if compressed

    :param input_path: List of nifti files path
    :param mask: path of mask or array
    :param output_flattened_array_path: path of the output flattened array. No extension is needed. Will be save as
        .npy if no compression, else .npz
    :param save: save the flattened array
    :param dtype: dtype of the output flattened array. Default is float 16 to save memory
    :param compress_save: If true compress the numpy array into .npz
    :return: flattened array [n_images, n_masked_voxels]
    """
    if isinstance(mask, str):
        mask, _ = load_nifty_volume_as_array(input_path_file=mask)

    if not isinstance(dtype(1), np.floating):
        raise ValueError("dtype need to be float type")

    logical_mask = mask == 1  # force the mask to be logical type
    n_voxels_flattened = np.sum(logical_mask)
    n_images = len(input_path)

    required_memory = mat_to_bytes(
        nrows=n_images,
        ncols=n_voxels_flattened,
        dtype=int(re.findall("\d+", np.dtype(dtype).name)[0]),
        out="MB",
    )
    print(f"required memory in RAM is: {required_memory:.2f} MB")

    flattened_array = np.zeros((n_images, n_voxels_flattened)).astype(dtype)
    for i, image_path in enumerate(tqdm(input_path, desc="Flattened array")):
        image_arr, _ = load_nifty_volume_as_array(image_path)
        flattened_array[i, :] = image_arr[logical_mask]

    if save:
        if not os.path.exists(os.path.dirname(output_flattened_array_path)):
            os.makedirs(os.path.dirname(output_flattened_array_path), exist_ok=True)
        if compress_save:
            np.savez_compressed(output_flattened_array_path, flattened_array)
        else:
            np.save(output_flattened_array_path, flattened_array)

    return flattened_array



import torch
import torch.nn as nn

class VascularAttention3D(nn.Module):
    """
    Vascular Attention Module for 3D CNNs.
    Learns spatial attention weights to focus on vascular regions.
    """
    def __init__(self, in_channels):
        super().__init__()
        self.attention_conv = nn.Conv3d(in_channels, 1, kernel_size=1) # 1x1 conv to learn attention map
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        """
        Forward pass of the Vascular Attention Module.

        Args:
            x (torch.Tensor): Input feature map of shape (B, C, D, H, W).

        Returns:
            torch.Tensor: Attention-weighted feature map of shape (B, C, D, H, W).
        """
        attention_map = self.sigmoid(self.attention_conv(x)) # (B, 1, D, H, W) - spatial attention map
        return x * attention_map # Apply attention weights to input features


import logging
import textwrap

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import wandb


def wrap_title(title, width=80):
    return textwrap.fill(title, width)

def get_normalization(norm_type, num_features, num_groups=8):
    """
    Returns a normalization layer based on the given type.
    Options:
      - 'batch': BatchNorm3d
      - 'group': GroupNorm (with num_groups groups)
      - 'instance': InstanceNorm3d
    """
    if norm_type == "batch":
        return nn.BatchNorm3d(num_features)
    elif norm_type == "group":
        # for GroupNorm, the number of groups is adjustable.
        return nn.GroupNorm(num_groups, num_features)
    elif norm_type == "instance":
        return nn.InstanceNorm3d(num_features)
    else:
        raise ValueError("Normalization type not recognized. Use 'batch', 'group', or 'instance'.")


def create_demographics_table(dataset, wandb_run):
    """Creates and logs a demographic table to wandb."""
    df = dataset.data_df.copy()
    # group by 'Site', 'Sex' and calculate the mean and std for 'Age'
    summary_df = (
        df.groupby(["Site", "Sex"])
        .agg(
            mean_age=pd.NamedAgg(column="Age", aggfunc="mean"),
            std_age=pd.NamedAgg(column="Age", aggfunc="std"),
            count=pd.NamedAgg(column="Age", aggfunc="count"),
        )
        .reset_index()
    )
    # create a wandb table, could be adjusted later with more columns
    demographics_table = wandb.Table(dataframe=summary_df)
    wandb_run.log({"demographics_table": demographics_table})
    logging.info("Demographics table created and logged to wandb.")

def create_prediction_chart(model, test_loader, device, wandb_run):
    """Creates and logs a prediction vs actual chart to wandb."""
    model.eval()
    all_predictions = []
    all_targets = []
    with torch.no_grad():
        for batch in test_loader:
            images = batch["image"].unsqueeze(1).to(device)
            ages = batch["age"].unsqueeze(1).to(device)
            demographics = batch["demographics"].to(device)
            outputs = model(images, demographics)
            all_predictions.extend(outputs.cpu().numpy().flatten())
            all_targets.extend(ages.cpu().numpy().flatten())
    plt.figure(figsize=(8, 6))
    plt.scatter(all_targets, all_predictions, alpha=0.5)
    plt.xlabel("Actual Age")
    plt.ylabel("Predicted Age")
    plt.title("Actual vs Predicted Age")
    plt.plot(
        [min(all_targets), max(all_targets)],
        [min(all_targets), max(all_targets)],
        color="red",
    )  # diagonal line fixed
    plt.grid(True)
    plt.tight_layout()
    wandb_run.log({"prediction_chart": plt})
    logging.info("Prediction chart created and logged to wandb.")


import torch
import torch.nn as nn
from .resnext3d import SEBlock3D

class DenseBlock3D(nn.Module):
    """
    DenseBlock3D with adjustable SE block.
    """
    def __init__(self, in_channels, growth_rate, num_layers, use_se=False):
        super(DenseBlock3D, self).__init__()
        self.layers = nn.ModuleList()
        current_channels = in_channels
        for _ in range(num_layers):
            layer = nn.Sequential(
                nn.BatchNorm3d(current_channels),
                nn.ReLU(inplace=True),
                nn.Conv3d(current_channels, growth_rate, kernel_size=3, padding=1),
            )
            if use_se:
                layer = nn.Sequential(layer, SEBlock3D(growth_rate))
            self.layers.append(layer)
            current_channels += growth_rate

    def forward(self, x):
        features = [x]
        for layer in self.layers:
            new_features = layer(torch.cat(features, 1))
            features.append(new_features)
        return torch.cat(features, 1)


class DenseNet3D(nn.Module):
    """
    DenseNet3D with adjustable parameters: growth rate, number of layers in DenseBlocks, use of SE blocks, dropout.
    """
    def __init__(self, num_demographics, growth_rate=16, num_dense_layers=[4, 4], initial_filters=32, transition_filters_multiplier=2, use_se_blocks=False, use_dropout=True, use_demographics=False):
        super(DenseNet3D, self).__init__()
        self.growth_rate = growth_rate
        self.num_dense_layers = num_dense_layers
        self.initial_filters = initial_filters
        self.transition_filters_multiplier = transition_filters_multiplier
        self.use_se_blocks = use_se_blocks
        self.use_dropout = use_dropout
        self.use_demographics = use_demographics

        self.conv1 = nn.Conv3d(1, initial_filters, kernel_size=7, stride=2, padding=3)
        self.bn1 = nn.BatchNorm3d(initial_filters)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool3d(kernel_size=3, stride=2, padding=1)

        current_filters = initial_filters
        self.dense1 = DenseBlock3D(current_filters, growth_rate, num_dense_layers[0], use_se=use_se_blocks)
        current_filters += num_dense_layers[0] * growth_rate
        next_filters = int(current_filters * transition_filters_multiplier)
        self.trans1 = nn.Sequential(
            nn.BatchNorm3d(current_filters),
            nn.Conv3d(current_filters, next_filters, kernel_size=1),
            nn.AvgPool3d(kernel_size=2, stride=2),
        )
        current_filters = next_filters

        self.dense2 = DenseBlock3D(current_filters, growth_rate, num_dense_layers[1], use_se=use_se_blocks)
        current_filters += num_dense_layers[1] * growth_rate
        next_filters = int(current_filters * transition_filters_multiplier)
        self.trans2 = nn.Sequential(
            nn.BatchNorm3d(current_filters),
            nn.Conv3d(current_filters, next_filters, kernel_size=1),
            nn.AvgPool3d(kernel_size=2, stride=2),
        )
        current_filters = next_filters


        self.avgpool = nn.AdaptiveAvgPool3d((1, 1, 1))
        self.fc1 = nn.Linear(current_filters, 64)
        self.dropout_layer = nn.Dropout(0.5) if use_dropout else nn.Identity()
        fc2_input_size = 64 + num_demographics if self.use_demographics else 64
        self.fc2 = nn.Linear(fc2_input_size, 1)

    def forward(self, x, demographics):
        """Forward pass."""
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.dense1(x)
        x = self.trans1(x)
        x = self.dense2(x)
        x = self.trans2(x)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = self.relu(x)
        x = self.dropout_layer(x)
        if self.use_demographics:
            x = torch.cat((x, demographics), dim=1)
        x = self.fc2(x)
        return x

    def get_name(self):
        """Dynamically generate model name based on parameters."""
        name = "DenseNet3D"
        name += f"_layers{'-'.join(map(str, self.num_dense_layers))}_gr{self.growth_rate}_filters{self.initial_filters}_transmult{self.transition_filters_multiplier}"
        if self.use_se_blocks:
            name += "_SE"
        if self.use_dropout:
            name += "_DO"
        if self.use_demographics:
            name += "_with_demographics"
        else:
            name += "_without_demographics"
        return name

    def get_params(self):
        """Return model parameters for wandb config."""
        return {
            "growth_rate": self.growth_rate,
            "num_dense_layers": self.num_dense_layers,
            "initial_filters": self.initial_filters,
            "transition_filters_multiplier": self.transition_filters_multiplier,
            "use_se_blocks": self.use_se_blocks,
            "use_dropout": self.use_dropout,
            "use_demographics": self.use_demographics,
            "architecture": "DenseNet3D"
        }


import argparse
import datetime
import logging
import os

import numpy as np
import pandas as pd
import torch
import torch.optim as optim
import wandb
from scipy.stats import pearsonr
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from sklearn.model_selection import GroupShuffleSplit, train_test_split
from torch.cuda.amp import GradScaler, autocast
from torch.utils.data import DataLoader

from .data import BrainAgeDataset
from .models.cnn import Large3DCNN
from .models.densenet3d import DenseNet3D
from .models.efficientnet3d import EfficientNet3D
from .models.improvedcnn3d import Improved3DCNN
from .models.loss import BrainAgeLoss
from .models.resnet3d import ResNet3D
from .models.resnext3d import ResNeXt3D
from .models.vit3d import VisionTransformer3D
from .utils import create_demographics_table, create_prediction_chart

torch.backends.cudnn.benchmark = True
torch.set_float32_matmul_precision("high")

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)

torch.manual_seed(42)
np.random.seed(42)

def verify_brain_masking(dataset, output_dir=None, num_samples=5):
    """
    Verifies brain masking by plotting the middle slice in three orthogonal views 
    for the first few images in the dataset immediately after they're loaded.
    
    Args:
        dataset: BrainAgeDataset or list of samples
        output_dir: Directory to save the verification plots (defaults to './mask_verification')
        num_samples: Number of samples to check
    """
    import os

    import matplotlib.pyplot as plt
    import numpy as np
    from matplotlib.colors import ListedColormap
    
    if output_dir is None:
        output_dir = './mask_verification'
    os.makedirs(output_dir, exist_ok=True)
    
    view_names = ['sagittal', 'coronal', 'axial']
    view_axes = {'sagittal': 0, 'coronal': 1, 'axial': 2}
    
    # Create a custom colormap with true black for zeros
    custom_gray = plt.cm.gray(np.arange(256))
    custom_gray[0] = [0, 0, 0, 1]  # Pure black for zero values
    custom_cmap = ListedColormap(custom_gray)
    
    # Determine if we're dealing with a dataset or a list
    if hasattr(dataset, '__getitem__') and not isinstance(dataset, list):
        sample_getter = lambda i: dataset[i]
        n_samples = min(num_samples, len(dataset))
    else:
        sample_getter = lambda i: dataset[i]
        n_samples = min(num_samples, len(dataset))
    
    for i in range(n_samples):
        sample = sample_getter(i)
        if sample is None:
            logging.info(f"Sample {i} is None, skipping")
            continue
            
        image = sample['image'].numpy() if torch.is_tensor(sample['image']) else sample['image']
        age = sample['age'].item() if torch.is_tensor(sample['age']) else sample['age']
        patient_id = sample.get('patient_id', f"patient_{i}")
        
        if isinstance(patient_id, torch.Tensor):
            patient_id = patient_id.item() if patient_id.numel() == 1 else str(patient_id)
        
        fig, axes = plt.subplots(4, 3, figsize=(15, 20))
        fig.suptitle(f"Patient {patient_id} - Age {age:.1f} - Mask Verification", fontsize=16)
        
        # Calculate masking statistics
        total_voxels = image.size
        zero_voxels = np.sum(image == 0)
        nonzero_voxels = total_voxels - zero_voxels
        percent_zero = (zero_voxels / total_voxels) * 100
        percent_nonzero = 100 - percent_zero
        
        stats_text = (
            f"Total voxels: {total_voxels:,}\n"
            f"Zero voxels: {zero_voxels:,} ({percent_zero:.2f}%)\n"
            f"Non-zero voxels: {nonzero_voxels:,} ({percent_nonzero:.2f}%)"
        )
        fig.text(0.5, 0.02, stats_text, ha='center', fontsize=12, bbox=dict(facecolor='white', alpha=0.5))
        
        for j, view_name in enumerate(view_names):
            view_axis = view_axes[view_name]
            slice_index = image.shape[view_axis] // 2
            
            # Get middle slice
            original_slice = np.take(image, indices=slice_index, axis=view_axis)
            
            # Create binary mask (1 where brain is present, 0 elsewhere)
            binary_mask = original_slice != 0
            
            # Row 1: Original image with default colormap (shows how matplotlib renders by default)
            axes[0, j].imshow(original_slice, cmap='gray')
            axes[0, j].set_title(f"{view_name.capitalize()} - Default gray colormap")
            axes[0, j].axis('off')
            
            # Row 2: Original image with custom colormap (pure black for zeros)
            axes[1, j].imshow(original_slice, cmap=custom_cmap, vmin=0)
            axes[1, j].set_title(f"{view_name.capitalize()} - Custom colormap (black zeros)")
            axes[1, j].axis('off')
            
            # Row 3: Binary mask (white = brain, black = background)
            axes[2, j].imshow(binary_mask, cmap='binary')
            axes[2, j].set_title(f"{view_name.capitalize()} - Binary mask")
            axes[2, j].axis('off')
            
            # Row 4: Histogram of pixel values
            if np.any(binary_mask):
                # Log-scale histogram of non-zero values
                nonzero_values = original_slice[binary_mask]
                axes[3, j].hist(nonzero_values.flatten(), bins=50, log=True)
                axes[3, j].axvline(x=0, color='r', linestyle='--', alpha=0.7)
                axes[3, j].set_title(f"{view_name.capitalize()} - Histogram (log scale)")
            else:
                axes[3, j].text(0.5, 0.5, "No non-zero values", ha='center')
                axes[3, j].set_title(f"{view_name.capitalize()} - No brain voxels")
                axes[3, j].axis('off')
        
        plt.tight_layout(rect=[0, 0.04, 1, 0.96])
        plt.savefig(os.path.join(output_dir, f"mask_verification_patient_{patient_id}.png"), dpi=300, bbox_inches='tight')
        plt.close(fig)
        
        # Create a mask boundary visualization
        fig2, axes2 = plt.subplots(1, 3, figsize=(15, 5))
        fig2.suptitle(f"Patient {patient_id} - Mask Boundary Visualization", fontsize=16)
        
        for j, view_name in enumerate(view_names):
            view_axis = view_axes[view_name]
            slice_index = image.shape[view_axis] // 2
            
            # Get original slice
            original_slice = np.take(image, indices=slice_index, axis=view_axis)
            brain_mask = original_slice != 0
            
            # Use masked array for proper visualization
            masked_data = np.ma.masked_where(~brain_mask, original_slice)
            
            # Plot with proper masking
            axes2[j].imshow(original_slice, cmap='gray', alpha=0.7)
            
            # Highlight the boundary
            from scipy import ndimage
            boundary = ndimage.binary_dilation(brain_mask, iterations=1) ^ brain_mask
            y, x = np.where(boundary)
            if len(y) > 0:
                axes2[j].scatter(x, y, s=1, c='red', alpha=0.8)
                
            axes2[j].set_title(f"{view_name.capitalize()} - Mask Boundary")
            axes2[j].axis('off')
        
        plt.tight_layout()
        plt.savefig(os.path.join(output_dir, f"mask_boundary_{patient_id}.png"), dpi=300, bbox_inches='tight')
        plt.close(fig2)
    
    logging.info(f"Saved mask verification images for {n_samples} patients to: {output_dir}")
    return output_dir


def train_model(
    csv_file,
    image_dir,
    model,
    model_type="large", 
    batch_size=4,
    learning_rate=0.001,
    num_epochs=100,
    use_wandb=True,
    pretrained_model_path=None,
    use_cuda=False,
    split_strategy="stratified_group",
    test_size=0.2,
    bins=10,
    output_dir="./saved_models",
    wandb_prefix="",
    brainage_alpha=0.0,
    brainage_beta=0.0,
    brainage_gamma=0.0,
    brainage_eps=1e-8,
    brainage_smoothing=0.0,
    brainage_use_huber=False,
    brainage_delta=0.0,
    weight_decay=0.05,
    store_model=True,
    mem_opt=False,
):
    logging.info("Starting training process...")
    os.makedirs(output_dir, exist_ok=True)

    
    try:
        model_name = model.get_name()
    except AttributeError:
        model_name = model_type 
    scaler = GradScaler()
    lr_str = f"{learning_rate:.1e}".replace("+", "").replace("-", "_")
    param_str = (
        f"{wandb_prefix}_{model_name}_{lr_str}_{num_epochs}_" 
        f"{batch_size}_{brainage_alpha}_{brainage_beta}_{brainage_gamma}_{brainage_eps}_{brainage_smoothing}_{brainage_use_huber}_{brainage_delta}"
    )

    if use_wandb:
        wandb_config = {
            "model_type": model_type,
            "batch_size": batch_size,
            "learning_rate": learning_rate,
            "num_epochs": num_epochs,
            "use_cuda": use_cuda,
            "split_strategy": split_strategy,
            "test_size": test_size,
            "bins": bins,
            "weight_decay": weight_decay,
            "brainage_alpha": brainage_alpha,
            "brainage_beta": brainage_beta,
            "brainage_gamma": brainage_gamma,
            "brainage_eps": brainage_eps,
            "brainage_smoothing": brainage_smoothing,
            "brainage_use_huber": brainage_use_huber,
            "brainage_delta": brainage_delta,
            "params": param_str,
            
        }
        try:
            model_params = model.get_params()
            wandb_config.update(model_params)
        except AttributeError:
            logging.info("Model does not have get_params method, skipping parameter logging to wandb.")

        wandb.init(
            project="asl-brainage",
            name=param_str,
            config=wandb_config,
        )
        run = wandb.run
    else:
        run = None
    dataset = BrainAgeDataset(csv_file, image_dir,mask_path=csv_file)
    if run:
        create_demographics_table(dataset, run)
    dataset = [
        sample for sample in dataset if sample is not None
    ]
    logging.info(
        "Number of samples after filtering for missing data: %d", len(dataset))
    device = torch.device(
        "cuda" if torch.cuda.is_available() and use_cuda else "cpu")
    logging.info("Using device: %s", device)
    if device.type == "cuda":
        gpu_name = torch.cuda.get_device_name(0)
        gpu_memory = (
            torch.cuda.get_device_properties(0).total_memory / 1024**3
        )
        gpu_info = (
            f"**\033[1mDetected GPU:\033[0m** {gpu_name}, Memory: {gpu_memory:.2f} GB"
        )
        logging.info(gpu_info)

    train_size = int(0.8 * len(dataset))
    test_size = len(dataset) - train_size

    logging.info(f"Using split strategy: {split_strategy}")

    if split_strategy == "random":
        train_size = int((1 - test_size) * len(dataset))
        test_size = len(dataset) - train_size
        train_dataset, test_dataset = torch.utils.data.random_split(
            dataset, [train_size, test_size]
        )
        
    elif split_strategy == "stratified":
        ages = [sample["age"].item() for sample in dataset]
        age_bins = pd.qcut(ages, q=bins, labels=False)
        train_idx, test_idx = train_test_split(
            range(len(dataset)), test_size=test_size, stratify=age_bins, random_state=42
        )
        train_dataset = torch.utils.data.Subset(dataset, train_idx)
        test_dataset = torch.utils.data.Subset(dataset, test_idx)
    elif split_strategy == "group":
        sites = [
            sample["demographics"][1].item() for sample in dataset
        ]  # Site is index 1
        gss = GroupShuffleSplit(
            n_splits=1, test_size=test_size, random_state=42)
        train_idx, test_idx = next(
            gss.split(range(len(dataset)), groups=sites))
        train_dataset = torch.utils.data.Subset(dataset, train_idx)
        test_dataset = torch.utils.data.Subset(dataset, test_idx)
    elif split_strategy == "stratified_group_sex":
        # Combine age bins and site for stratification
        ages = [sample["age"].item() for sample in dataset]
        sites = [sample["demographics"][1].item() for sample in dataset]
        sexes = [
            sample["demographics"][0].item() for sample in dataset
        ]  
        age_bins = pd.qcut(ages, q=bins, labels=False)
        stratify_groups = [
            f"{site}_{sex}_{age_bin}"
            for site, sex, age_bin in zip(sites, sexes, age_bins)
        ]  
        train_idx, test_idx = train_test_split(
            range(len(dataset)),
            test_size=test_size,
            stratify=stratify_groups,
            random_state=42,
        )
        train_dataset = torch.utils.data.Subset(dataset, train_idx)
        test_dataset = torch.utils.data.Subset(dataset, test_idx)
    elif split_strategy == "stratified_group":
        
        ages = [sample["age"].item() for sample in dataset]
        sites = [sample["demographics"][1].item() for sample in dataset]
        age_bins = pd.qcut(ages, q=bins, labels=False)
        stratify_groups = [
            f"{site}_{age_bin}" for site, age_bin in zip(sites, age_bins)
        ]
        train_idx, test_idx = train_test_split(
            range(len(dataset)),
            test_size=test_size,
            stratify=stratify_groups,
            random_state=42,
        )
        train_dataset = torch.utils.data.Subset(dataset, train_idx)
        test_dataset = torch.utils.data.Subset(dataset, test_idx)
    else:
        raise ValueError(f"Invalid split strategy: {split_strategy}")
    #save indices of test dataset
    
    csv_file_path = os.path.dirname(csv_file)
    if os.path.exists(os.path.join(csv_file_path, f"train_indices.npy")) and os.path.exists(os.path.join(csv_file_path, f"test_indices.npy")):
        train_idx = np.load(os.path.join(csv_file_path, f"train_indices.npy"))
        test_idx = np.load(os.path.join(csv_file_path, f"test_indices.npy"))
        train_dataset = torch.utils.data.Subset(dataset, train_idx)
        test_dataset = torch.utils.data.Subset(dataset, test_idx)
        logging.info("Train and test indices loaded")
    else:        
        np.save(os.path.join(csv_file_path, f"train_indices.npy"), train_idx)
        np.save(os.path.join(csv_file_path, f"test_indices.npy"), test_idx)
        logging.info("Train and test indices saved")

    train_loader = DataLoader(
        train_dataset, batch_size=batch_size, shuffle=True, num_workers = 2, pin_memory=True)
    test_loader = DataLoader(
        test_dataset, batch_size=batch_size, shuffle=False, num_workers = 2, pin_memory=True)
    
    logging.info("Data loaders created")
    
    model.to(device)
    cmodel = torch.compile(model)
    logging.info(f"Model created: {cmodel.get_name() if hasattr(cmodel, 'get_name') else model_type}")

    #criterion = nn.L1Loss()
    criterion = BrainAgeLoss(
        alpha=brainage_alpha,
        beta=brainage_beta,
        gamma=brainage_gamma,
        eps=brainage_eps,
        smoothing=brainage_smoothing,
        use_huber=brainage_use_huber,
        delta=brainage_delta
    )
    logging.info(f"Loss function: {criterion.get_name()} with parameters: alpha={brainage_alpha}, beta={brainage_beta}, gamma={brainage_gamma}, eps={brainage_eps}, smoothing={brainage_smoothing}, use_huber={brainage_use_huber}, delta={brainage_delta}")

    #optimizer = optim.Adam(cmodel.parameters(), lr=learning_rate, weight_decay=weight_decay)
    # scheduler = optim.lr_scheduler.CosineAnnealingLR.ReduceLROnPlateau(
    #     optimizer,
    #     mode='min',
    #     factor=0.5,
    #     patience=20,
    #     verbose=True,
    # )
    optimizer = torch.optim.AdamW(model.parameters(), lr=learning_rate, weight_decay=weight_decay)
    warmup_epochs = 20
    eta_min = 1e-6
    warmup_scheduler = optim.lr_scheduler.LinearLR(
    optimizer,
    start_factor=1e-6/learning_rate,  # start from a very small LR
    end_factor=1.0,              # reach initial_lr at the end of warmup
    total_iters=warmup_epochs    # number of warmup epochs
    )
    main_scheduler = optim.lr_scheduler.CosineAnnealingLR(
    optimizer,
    T_max=num_epochs - warmup_epochs,  # adjust T_max for warmup, this seems to be working fine for our use-case
    eta_min=eta_min
    )
    scheduler = optim.lr_scheduler.SequentialLR(
    optimizer,
    schedulers=[warmup_scheduler, main_scheduler],
    milestones=[warmup_epochs]  # fallback to main_scheduler after warmup
    )
    
    logging.info(f"Loss function and optimizer set up.")
    current_time = datetime.datetime.now().strftime("%y-%m-%d_%H-%M")
    best_model_path = os.path.join(output_dir, f"{param_str}_{current_time}.pth")
    if not pretrained_model_path:
        best_test_mae = float("inf")
        epochs_no_improve = 0
        for epoch in range(num_epochs):
            logging.info(f"Starting epoch: {epoch}")
            cmodel.train()
            train_loss = 0
            for i, batch in enumerate(train_loader):
                logging.debug(f"Processing batch: {i}")
                images = (
                    batch["image"].unsqueeze(1).to(device)
                )
                ages = batch["age"].unsqueeze(1).to(device)
                demographics = batch["demographics"].to(device)
                optimizer.zero_grad()
                if mem_opt:
                    with autocast():
                        outputs = cmodel(images, demographics)
                        loss, metrics = criterion(outputs, ages, demographics)
                    scaler.scale(loss).backward()
                    scaler.step(optimizer)
                    scaler.update()
                else:
                    outputs = cmodel(images, demographics)
                    #loss = criterion(outputs, ages)
                    loss, metrics = criterion(outputs, ages, demographics)
                    loss.backward()
                    optimizer.step()
                train_loss += loss.item()
                logging.debug(f"Batch {i} processed. Loss: {loss.item()}")
            train_loss = train_loss / len(train_loader)
            logging.info(f"Epoch: {epoch}, Training Loss: {train_loss}")
            cmodel.eval()
            with torch.no_grad():
                test_mae, test_rmse, test_r2, test_pearson, test_mape = evaluate_model(
                    cmodel, test_loader, device
                )
            logging.info(
                f"Epoch {epoch} Test MAE: {test_mae}, Test MAPE: {test_mape}, RMSE: {test_rmse}, R2: {test_r2}, Pearson Correlation: {test_pearson}"
            )
            if run:
                wandb.log(
                    {
                        "train_loss": train_loss,
                        "test_mae": test_mae,
                        "test_mape": test_mape,
                        "test_rmse": test_rmse,
                        "test_r2": test_r2,
                        "test_pearson": test_pearson,
                        "metrics_mae": metrics["mae"],
                        "metrics_corr_loss": metrics["corr_loss"],
                        "metrics_std_ratio_loss": metrics["std_ratio_loss"],
                        "metrics_correlation": metrics["correlation"],
                        "metrics_pred_std": metrics["pred_std"],
                        "metrics_target_std": metrics["target_std"],
                        
                    }
                )
            #scheduler.step(test_mae)
            scheduler.step()
            #log if learning rate changes
            logging.info(f"-> Learning rate: {optimizer.param_groups[0]['lr']}") if optimizer.param_groups[0]['lr'] != learning_rate else None
            if round(test_mae, 2) < round(best_test_mae, 2):
                epochs_no_improve = 0
                best_test_mae = test_mae
                if store_model:
                    _sd = cmodel.state_dict()
                    _sd = {key.replace("_orig_mod.", ""): value for key, value in _sd.items()}
                    model.load_state_dict(_sd)
                    torch.save(model, best_model_path)

                    logging.info(
                        f"Model saved at epoch {epoch} as test MAE {test_mae} is better than previous best {best_test_mae}"
                    )
            else:
                epochs_no_improve += 1
                if epochs_no_improve >= 40:
                    logging.info(
                        f"Early stopping at epoch {epoch} as test MAE did not improve over 20 epochs"
                    )
                    break

        logging.info("Training completed.")
    cmodel.eval()
    with torch.no_grad():
        test_mae, test_rmse, test_r2, test_pearson, test_mape = evaluate_model(
            cmodel, test_loader, device
        )
    logging.info(
        f"Final Test MAE: {test_mae}, Test MAPE: {test_mape}, RMSE: {test_rmse}, R2: {test_r2}, Pearson Correlation: {test_pearson}"
    )
    if run:
        create_prediction_chart(cmodel, test_loader, device, run)
        wandb.log(
            {
                "final_test_mae": test_mae,
                "final_test_mape": test_mape,
                "final_test_rmse": test_rmse,
                "final_test_r2": test_r2,
                "final_test_pearson": test_pearson,
            }
        )
    if run:
        wandb.finish()
    return model


def evaluate_model(model, test_loader, device):
    model.eval()
    all_predictions = []
    all_targets = []
    with torch.no_grad():
        for batch in test_loader:
            images = batch["image"].unsqueeze(1).to(device)
            ages = batch["age"].unsqueeze(1).to(device)
            demographics = batch["demographics"].to(device)
            outputs = model(images, demographics)  # forward pass
            all_predictions.extend(
                outputs.cpu().numpy().flatten()
            )
            all_targets.extend(
                ages.cpu().numpy().flatten()
            )
    all_predictions = np.array(all_predictions)
    all_targets = np.array(all_targets)
    mae = mean_absolute_error(all_targets, all_predictions)
    rmse = np.sqrt(mean_squared_error(
        all_targets, all_predictions))

    r2 = r2_score(all_targets, all_predictions)
    pearson, _ = pearsonr(all_targets, all_predictions)
    mape = np.mean(np.abs((all_targets - all_predictions) / (all_targets + 1e-8))) * 100
    return mae, rmse, r2, pearson, mape


def main():
    parser = argparse.ArgumentParser(
        description="Brain Age Prediction Training Script")

    parser.add_argument(
        "--csv_file",
        type=str,
        default="/home/radv/samiri/my-scratch/trainingdata/topmri.csv",
        help="Path to the CSV file",
    )
    parser.add_argument(
        "--image_dir",
        type=str,
        default="/home/radv/samiri/my-scratch/trainingdata/topmri/",
        help="Path to the image directory",
    )
    parser.add_argument("--wandb_prefix", type=str,
                        default="", help="wandb job prefix")
    parser.add_argument(
        "--batch_size", type=int, default=4, help="Batch size for training"
    )
    parser.add_argument(
        "--learning_rate",
        type=float,
        default=0.000015,
        help="Learning rate for training",
    )
    parser.add_argument(
        "--num_epochs", type=int, default=150, help="Number of epochs for training"
    )
    parser.add_argument("--use_wandb", action="store_true",
                        help="Enable wandb logging")
    parser.add_argument(
        "--pretrained_model_path",
        type=str,
        default=None,
        help="Path to a pretrained model file",
    )
    parser.add_argument(
        "--use_cuda",
        action="store_true",
        default=True,
        help="Enable CUDA (GPU) if available",
    )
    parser.add_argument(
      "--store_model",
      action="store_true",
      help="Store the best model",
    )

    parser.add_argument(
        "--split_strategy",
        type=str,
        default="stratified_group_sex",
        choices=[
            "random",
            "stratified",
            "group",
            "stratified_group",
            "stratified_group_sex",
        ],
        help="Data split strategy (random/stratified/group/stratified_group)",
    )
    parser.add_argument(
        "--weight_decay",
        type=float,
        default=0.0,
        help="Weight decay for Adam optimizer",
    )
    parser.add_argument(
        "--test_size",
        type=float,
        default=0.2,
        help="Proportion of data to use for testing",
    )
    parser.add_argument(
        "--bins", type=int, default=10, help="Number of bins for stratification"
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default="./saved_models",
        help="Directory to save trained models",
    )

    parser.add_argument(
        "--model_type",
        type=str,
        default="large",
        choices=[
            "large",
            "resnet",
            "densenet",
            "resnext3d",
            "efficientnet3d",
            "improved_cnn",
            "vit3d",
        ],
        help="Type of model to use",
    )


    # Large3DCNN Parameters
    large_cnn_group = parser.add_argument_group("Large3DCNN arguments")
    large_cnn_group.add_argument("--large_cnn_layers", type=int, default=3, help="Number of conv layers for Large3DCNN")
    large_cnn_group.add_argument("--large_cnn_filters", type=int, default=16, help="Initial filters for Large3DCNN")
    large_cnn_group.add_argument("--large_cnn_filters_multiplier", type=float, default=2.0, help="Filters multiplier for Large3DCNN") # Added filter multiplier
    large_cnn_group.add_argument("--large_cnn_use_bn", action="store_true", help="Use BN in Large3DCNN")
    large_cnn_group.add_argument("--large_cnn_use_se", action="store_true", help="Use SE blocks in Large3DCNN")
    large_cnn_group.add_argument("--large_cnn_use_dropout", action="store_true", help="Use dropout in Large3DCNN")
    large_cnn_group.add_argument("--large_cnn_dropout_rate", type=float, default=0.2, help="Dropout rate for Large3DCNN")

    # ResNet3D Parameters
    resnet_group = parser.add_argument_group("ResNet3D arguments")
    resnet_group.add_argument("--resnet_layers", type=int, nargs='+', default=[2, 2, 2], help="Number of blocks per ResNet layer (list of 3 ints)")
    resnet_group.add_argument("--resnet_initial_filters", type=int, default=32, help="Initial filters for ResNet3D")
    resnet_group.add_argument("--resnet_filters_multiplier", type=float, default=2.0, help="Filters multiplier for ResNet3D") # Added filter multiplier
    resnet_group.add_argument("--resnet_use_se", action="store_true", help="Use SE blocks in ResNet3D")
    resnet_group.add_argument("--resnet_dropout", action="store_true", help="Use dropout in ResNet3D")

    # DenseNet3D Parameters
    densenet_group = parser.add_argument_group("DenseNet3D arguments")
    densenet_group.add_argument("--densenet_growth_rate", type=int, default=16, help="Growth rate for DenseNet3D")
    densenet_group.add_argument("--densenet_layers", type=int, nargs='+', default=[4, 4], help="Number of layers per DenseBlock (list of 2 ints)")
    densenet_group.add_argument("--densenet_initial_filters", type=int, default=32, help="Initial filters for DenseNet3D")
    densenet_group.add_argument("--densenet_transition_filters_multiplier", type=float, default=2.0, help="Filters multiplier for DenseNet3D transition layers") # Added filter multiplier
    densenet_group.add_argument("--densenet_use_se", action="store_true", help="Use SE blocks in DenseNet3D")
    densenet_group.add_argument("--densenet_dropout", action="store_true", help="Use dropout in DenseNet3D")

    # ResNeXt3D Parameters
    resnext_group = parser.add_argument_group("ResNeXt3D arguments")
    resnext_group.add_argument("--resnext_layers", type=int, nargs='+', default=[2, 2, 2], help="Number of blocks per ResNeXt layer (list of 3 ints)")
    resnext_group.add_argument("--resnext_cardinality", type=int, default=32, help="Cardinality for ResNeXt3D")
    resnext_group.add_argument("--resnext_bottleneck_width", type=int, default=4, help="Bottleneck width for ResNeXt3D")
    resnext_group.add_argument("--resnext_initial_filters", type=int, default=64, help="Initial filters for ResNeXt3D")
    resnext_group.add_argument("--resnext_filters_multiplier", type=float, default=2.0, help="Filters multiplier for ResNeXt3D") # Added filter multiplier
    resnext_group.add_argument("--resnext_use_se", action="store_true", help="Use SE blocks in ResNeXt3D")
    resnext_group.add_argument("--resnext_dropout", action="store_true", help="Use dropout in ResNeXt3D")

    # EfficientNet3D Parameters
    efficientnet_group = parser.add_argument_group("EfficientNet3D arguments")
    efficientnet_group.add_argument("--efficientnet_width_coefficient", type=float, default=1.0, help="Width coefficient for EfficientNet3D")
    efficientnet_group.add_argument("--efficientnet_depth_coefficient", type=float, default=1.0, help="Depth coefficient for EfficientNet3D")
    efficientnet_group.add_argument("--efficientnet_initial_filters", type=int, default=32, help="Initial filters for EfficientNet3D")
    efficientnet_group.add_argument("--efficientnet_filters_multiplier", type=float, default=1.2, help="Filters multiplier for EfficientNet3D") # Added filter multiplier
    efficientnet_group.add_argument("--efficientnet_dropout", type=float, default=0.2, help="Dropout rate for EfficientNet3D (set 0 to disable)")

    # Improved3DCNN Parameters
    improved_cnn_group = parser.add_argument_group("Improved3DCNN arguments")
    improved_cnn_group.add_argument("--improved_cnn_initial_filters", type=int, default=32, help="Initial filters for Improved3DCNN")
    improved_cnn_group.add_argument("--improved_cnn_filters_multiplier", type=float, default=2.0, help="Filters multiplier for Improved3DCNN") # Added filter multiplier
    improved_cnn_group.add_argument("--improved_cnn_num_conv_layers", type=int, default=3, help="Number of conv layers for Improved3DCNN") # Added num_conv_layers
    improved_cnn_group.add_argument("--improved_cnn_use_se", action="store_true", help="Use SE blocks in Improved3DCNN")
    improved_cnn_group.add_argument("--improved_cnn_dropout_rate", type=float, default=0.3, help="Dropout rate for Improved3DCNN")
    
    # VisionTransformer3D Parameters
    vit3d_group = parser.add_argument_group("VisionTransformer3D arguments")
    vit3d_group.add_argument("--vit3d_patch_size", type=int, nargs='+', default=[16, 16, 16], help="Patch size for ViT3D (list of 3 ints)")
    vit3d_group.add_argument("--vit3d_embed_dim", type=int, default=768, help="Embedding dimension for ViT3D")
    vit3d_group.add_argument("--vit3d_depth", type=int, default=12, help="Depth (number of layers) for ViT3D")
    vit3d_group.add_argument("--vit3d_num_heads", type=int, default=12, help="Number of attention heads for ViT3D")
    vit3d_group.add_argument("--vit3d_mlp_ratio", type=float, default=4.0, help="MLP ratio for ViT3D")
    vit3d_group.add_argument("--vit3d_qkv_bias", action="store_true", help="Use QKV bias in ViT3D attention")
    vit3d_group.add_argument("--vit3d_drop_rate", type=float, default=0.0, help="Dropout rate for ViT3D")
    vit3d_group.add_argument("--vit3d_attn_drop_rate", type=float, default=0.0, help="Attention dropout rate for ViT3D")
    vit3d_group.add_argument("--vit3d_drop_path_rate", type=float, default=0.0, help="Drop path rate for ViT3D")
    vit3d_group.add_argument("--vit3d_global_pool", action="store_true", help="Use global average pooling in ViT3D")    
    vit3d_group.add_argument("--vit3d_use_cls_token", action="store_true", help="Use CLS token in ViT3D")
    vit3d_group.add_argument("--vit3d_use_hybrid_embed", action="store_true", help="Use hybrid embedding in ViT3D")
    vit3d_group.add_argument("--vit3d_hybrid_kernel_size", type=int, default=3, help="Hybrid embedding kernel size for ViT3D")
    
    # BrainAgeLoss Parameters
    brainage_loss_group = parser.add_argument_group("BrainAgeLoss arguments")
    brainage_loss_group.add_argument("--brainage_alpha", type=float, default=1.0, help="Weight for correlation loss in BrainAgeLoss")
    brainage_loss_group.add_argument("--brainage_beta", type=float, default=0.3, help="Weight for bias regularization in BrainAgeLoss")
    brainage_loss_group.add_argument("--brainage_gamma", type=float, default=0.2, help="Weight for age-specific weighting in BrainAgeLoss")
    brainage_loss_group.add_argument("--brainage_eps", type=float, default=1e-8, help="Epsilon for numerical stability in BrainAgeLoss")
    brainage_loss_group.add_argument("--brainage_smoothing", type=float, default=0.1, help="Label smoothing factor in BrainAgeLoss")
    brainage_loss_group.add_argument("--brainage_use_huber", action="store_true", default=True, help="Use Huber loss in BrainAgeLoss")
    brainage_loss_group.add_argument("--brainage_delta", type=float, default=1.0, help="Delta parameter for Huber loss in BrainAgeLoss")
    

    args = parser.parse_args()

    num_demographics = 6  # Number of demographic features

    if args.model_type == "improved_cnn":
        model = Improved3DCNN(
            num_demographics=num_demographics,
            initial_filters=args.improved_cnn_initial_filters,
            filters_multiplier=args.improved_cnn_filters_multiplier,
            num_conv_layers=args.improved_cnn_num_conv_layers,
            use_se=args.improved_cnn_use_se,
            dropout_rate=args.improved_cnn_dropout_rate,
        )

    elif args.model_type == "large":
        model = Large3DCNN(
            num_demographics=num_demographics,
            num_conv_layers=args.large_cnn_layers,
            initial_filters=args.large_cnn_filters,
            filters_multiplier=args.large_cnn_filters_multiplier,
            use_bn=args.large_cnn_use_bn,
            use_se=args.large_cnn_use_se,
            use_dropout=args.large_cnn_use_dropout,
            dropout_rate=args.large_cnn_dropout_rate
        )

    elif args.model_type == "resnet":
        model = ResNet3D(
            num_demographics=num_demographics,
            num_blocks_per_layer=args.resnet_layers,
            initial_filters=args.resnet_initial_filters,
            filters_multiplier=args.resnet_filters_multiplier,
            use_se=args.resnet_use_se,
            use_dropout=args.resnet_dropout
        )

    elif args.model_type == "densenet":
        model = DenseNet3D(
            num_demographics=num_demographics,
            growth_rate=args.densenet_growth_rate,
            num_dense_layers=args.densenet_layers,
            initial_filters=args.densenet_initial_filters,
            transition_filters_multiplier=args.densenet_transition_filters_multiplier,
            use_se_blocks=args.densenet_use_se,
            use_dropout=args.densenet_dropout,
        )
    elif args.model_type == "resnext3d":
        model = ResNeXt3D(
            num_demographics=num_demographics,
            num_blocks_per_layer=args.resnext_layers,
            cardinality=args.resnext_cardinality,
            bottleneck_width=args.resnext_bottleneck_width,
            initial_filters=args.resnext_initial_filters,
            filters_multiplier=args.resnext_filters_multiplier,
            use_se=args.resnext_use_se,
            use_dropout=args.resnext_dropout,
        )
    elif args.model_type == "efficientnet3d":
        model = EfficientNet3D(
            num_demographics=num_demographics,
            width_coefficient=args.efficientnet_width_coefficient,
            depth_coefficient=args.efficientnet_depth_coefficient,
            initial_filters=args.efficientnet_initial_filters,
            filters_multiplier=args.efficientnet_filters_multiplier,
            use_dropout=args.efficientnet_dropout,
        )

    elif args.model_type == "vit3d":
        model = VisionTransformer3D(
            num_demographics=num_demographics,
            img_size=[121, 145, 121], 
            patch_size=args.vit3d_patch_size,
            embed_dim=args.vit3d_embed_dim,
            depth=args.vit3d_depth,
            num_heads=args.vit3d_num_heads,
            mlp_ratio=args.vit3d_mlp_ratio,
            qkv_bias=args.vit3d_qkv_bias,
            drop_rate=args.vit3d_drop_rate,
            attn_drop_rate=args.vit3d_attn_drop_rate,
            drop_path_rate=args.vit3d_drop_path_rate,
            global_pool=args.vit3d_global_pool,
            use_cls_token = args.vit3d_use_cls_token,
            use_hybrid_embed = args.vit3d_use_hybrid_embed,
            hybrid_kernel_size = args.vit3d_hybrid_kernel_size,
            use_demographics=True
        )
            
    else:
        raise ValueError(f"Invalid model_type: {args.model_type}")

    train_model(
        csv_file=args.csv_file,
        image_dir=args.image_dir,
        model=model,
        model_type=args.model_type,
        batch_size=args.batch_size,
        learning_rate=args.learning_rate,
        num_epochs=args.num_epochs,
        use_wandb=args.use_wandb,
        pretrained_model_path=args.pretrained_model_path,
        use_cuda=args.use_cuda,
        split_strategy=args.split_strategy,
        test_size=args.test_size,
        bins=args.bins,
        output_dir=args.output_dir,
        wandb_prefix=args.wandb_prefix,
        weight_decay=args.weight_decay,
        store_model=args.store_model,
        brainage_alpha=args.brainage_alpha,
        brainage_beta=args.brainage_beta,
        brainage_gamma=args.brainage_gamma,
        brainage_eps=args.brainage_eps,
        brainage_smoothing=args.brainage_smoothing,
        brainage_use_huber=args.brainage_use_huber,
        brainage_delta=args.brainage_delta,
        
    )

if __name__ == "__main__":
    main()